from sqlalchemy import Column, Integer, String, text, DateTime, Numeric, Text
from sqlalchemy.dialects.mysql import TIMESTAMP
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()
metadata = Base.metadata


class AuthDetect(Base):
    __tablename__ = 'tb_auth_detect'
    __table_args__ = {'comment': '原始交易授權驗證表'}

    pk_id = Column(Integer, primary_key=True,
                   autoincrement=True, nullable=False)
    client_info_id = Column(String(100, 'utf8mb4_unicode_ci'))
    sn = Column(String(60, 'utf8mb4_unicode_ci'), nullable=False, unique=True)
    customer_servertime = Column(TIMESTAMP(fsp=6), nullable=False)
    sp_tx_id = Column(String(45, 'utf8mb4_unicode_ci'))
    pan_hash = Column(String(128, 'utf8mb4_unicode_ci'), nullable=False)
    pan_expire = Column(String(6, 'utf8mb4_unicode_ci'), nullable=False)
    merchant_id = Column(String(45, 'utf8mb4_unicode_ci'), nullable=False)
    purchase_amount = Column(Numeric(16, 5), nullable=False)
    purchase_currency = Column(
        String(10, 'utf8mb4_unicode_ci'), nullable=False)
    eci = Column(String(3, 'utf8mb4_unicode_ci'), nullable=False)
    installment_period = Column(String(3, 'utf8mb4_unicode_ci'))
    redeem_flag = Column(String(3, 'utf8mb4_unicode_ci'))
    pay_way = Column(String(3, 'utf8mb4_unicode_ci'))
    cof_flag = Column(String(3, 'utf8mb4_unicode_ci'))
    acq_id = Column(String(16, 'utf8mb4_unicode_ci'))
    rtn_url = Column(Text, comment='付款結果回傳網址')
    card_bin = Column(String(20, 'utf8mb4_unicode_ci'))
    threeds_type = Column(String(3, 'utf8mb4_unicode_ci'), nullable=False)
    create_time = Column(TIMESTAMP(fsp=6), nullable=False,
                         server_default=text("CURRENT_TIMESTAMP(6)"))

    def __init__(self, **kwargs):
        for property, value in kwargs.items():
            setattr(self, property, value)


class AuthDetectCleaned(Base):
    __tablename__ = 'tb_auth_detect_cleaned'
    __table_args__ = {'comment': '原始交易授權驗證表(清洗後)'}

    pk_id = Column(Integer, primary_key=True,
                   autoincrement=True, nullable=False)
    client_info_id = Column(String(20, 'utf8mb4_unicode_ci'))
    sn = Column(String(60, 'utf8mb4_unicode_ci'), nullable=False, unique=True)
    servertime = Column(DateTime, nullable=False)
    sp_tx_id = Column(String(45, 'utf8mb4_unicode_ci'))
    hashcard = Column(String(128, 'utf8mb4_unicode_ci'), nullable=False)
    pan_expire = Column(Integer, nullable=False)
    merchant_id = Column(String(45, 'utf8mb4_unicode_ci'), nullable=False)
    purchase_amount = Column(Numeric(16, 5), nullable=False)
    purchase_currency = Column(
        String(45, 'utf8mb4_unicode_ci'), nullable=False)
    eci = Column(Integer, nullable=False)
    installment_period = Column(Integer)
    redeem_flag = Column(Integer)
    pay_way = Column(Integer)
    cof_flag = Column(Integer)
    bank_id = Column(String(16, 'utf8mb4_unicode_ci'))
    rtn_url = Column(Text, comment='付款結果回傳網址')
    card_bin = Column(String(20, 'utf8mb4_unicode_ci'))
    threeds_type = Column(Integer, nullable=False)
    create_time = Column(TIMESTAMP(fsp=6), nullable=False,
                         server_default=text("CURRENT_TIMESTAMP(6)"))


class AuthDetectAR(Base):
    __tablename__ = 'tb_auth_detect_ar'
    __table_args__ = {'comment': '原始交易授權驗證表(AR)'}

    pk_id = Column(Integer, primary_key=True,
                   autoincrement=True, nullable=False)
    client_info_id = Column(String(20, 'utf8mb4_unicode_ci'))
    sn = Column(String(60, 'utf8mb4_unicode_ci'), nullable=False, unique=True)
    timestamp = Column(DateTime, nullable=False)
    hashcard = Column(String(128, 'utf8mb4_unicode_ci'), nullable=False)
    sp_tx_id = Column(String(45, 'utf8mb4_unicode_ci'))
    pan_expire = Column(Integer, nullable=False)
    merchant_id = Column(String(45, 'utf8mb4_unicode_ci'), nullable=False)
    purchase_amount = Column(Numeric(16, 5), nullable=False)
    purchase_currency = Column(
        String(45, 'utf8mb4_unicode_ci'), nullable=False)
    eci = Column(Integer, nullable=False)
    installment_period = Column(Integer, nullable=False)
    redeem_flag = Column(Integer, nullable=False)
    pay_way = Column(Integer, nullable=False)
    cof_flag = Column(Integer, nullable=False)
    bank_id = Column(String(16, 'utf8mb4_unicode_ci'))
    rtn_url = Column(Text, comment='付款結果回傳網址')
    card_bin = Column(String(20, 'utf8mb4_unicode_ci'))
    threeds_type = Column(Integer, nullable=False)
    create_time = Column(TIMESTAMP(fsp=6), nullable=False,
                         server_default=text("CURRENT_TIMESTAMP(6)"))


class AuthDetectAI(Base):
    __tablename__ = 'tb_temp_auth_detect_ai'
    __table_args__ = {'comment': '原始交易授權驗證表(AI)'}

    pk_id = Column(Integer, primary_key=True,
                   autoincrement=True, nullable=False)
    client_info_id = Column(String(20, 'utf8mb4_unicode_ci'))
    sn = Column(String(60, 'utf8mb4_unicode_ci'), nullable=False, unique=True)
    timestamp = Column(DateTime, nullable=False)
    hashcard = Column(String(128, 'utf8mb4_unicode_ci'), nullable=False)
    ai = Column(Numeric(16, 5), nullable=False)
    cl = Column(Numeric(16, 5), nullable=False)
    ucl = Column(Numeric(16, 5), nullable=False)
    is_abnormal = Column(String(45, 'utf8mb4_unicode_ci'), nullable=False)
    create_time = Column(TIMESTAMP(fsp=6), nullable=False,
                         server_default=text("CURRENT_TIMESTAMP(6)"))


class AuthRealResult(Base):
    __tablename__ = "tb_auth_real_result"
    pk_id = Column(Integer, primary_key=True,
                   nullable=False, autoincrement=True)
    return_code = Column(String(10))
    client_info_id = Column(String(20, 'utf8mb4_unicode_ci'))
    sn = Column(String(60))
    customer_servertime = Column(DateTime, nullable=False)
    auth_return_code = Column(String(10))
    create_time = Column(TIMESTAMP(6), nullable=False)



import numpy as np


class AHPWeight:
    def __init__(self, ahp_weight):
        self.ahp_weight = ahp_weight

    def algorithm_ahp_scale(self, list_scale):
        def reciprocal(num):
            return [1. / np.array(abs(num))]

        dict_distance = {}
        # list_scale = [1, 2, -3, -4]
        list_array = []
        list_A_arr = []
        for i in list_scale:
            if i < 1:
                list_array.append(reciprocal(i)[0])
            else:
                list_array.append(i)
        # print(list_array)
        list_A_arr.append(list_array)

        for y in range(len(list_scale)):
            if y == 0:
                dict_distance[y] = list_array
            else:
                one_list = []
                for x in range(len(list_scale)):
                    # ====================正級距======================
                    if x == 0:
                        record = reciprocal(list_array[y])[0]
                        one_list.append(record)

                    if x != 0 and list_scale[y] > list_scale[x] and list_scale[x] > 0 and list_scale[y] > 0:
                        one_list.append(list_scale[x] / list_scale[y])
                    elif x != 0 and list_scale[y] == list_scale[x]:
                        one_list.append(1)
                    elif x != 0 and list_scale[y] < list_scale[x] and list_scale[y] > 0:
                        one_list.append(list_array[x] / list_array[y])
                    # y是正的，比較級距是負的
                    elif x != 0 and list_scale[y] > list_scale[x] and list_scale[x] < 0 and list_scale[y] > 0:
                        one_list.append(1 / (list_scale[y] - list_scale[x] + 1))

                    # ====================負級距======================
                    elif x != 0 and list_scale[y] > list_scale[x] and list_scale[y] < 0:
                        one_list.append((1 / list_scale[x]) / (1 / list_scale[y]))

                    elif x != 0 and list_scale[y] < list_scale[x] and list_scale[x] < 0 and list_scale[y] < 0:
                        one_list.append((1 / list_scale[x]) / (1 / list_scale[y]))

                    elif x != 0 and list_scale[y] < list_scale[x] and list_scale[x] > 0 and list_scale[y] < 0:
                        one_list.append(list_scale[x] - list_scale[y] + 1)

                list_A_arr.append(one_list)
        return list_A_arr

    def algorithm_ahp(self, A_arr):
        # 矩阵
        A = np.array(A_arr)
        a_sum0 = A.sum(axis=0)
        B = A / a_sum0
        b_sum = B.sum(axis=1)
        W = b_sum.sum()
        w_arr = []
        for w in b_sum:
            w_arr.append(w / W)
        return w_arr

    def output_weight(self):
        model_list, weight_list = list(self.ahp_weight.keys()), list(self.ahp_weight.values())
        idx = weight_list.index(1)

        weight_list[idx], weight_list[0] = weight_list[0], weight_list[idx]
        model_list[idx], model_list[0] = model_list[0], model_list[idx]
        ahp_weight = dict(zip(model_list, weight_list))

        list_scale = self.algorithm_ahp(self.algorithm_ahp_scale(weight_list))
        for i, (k, _) in enumerate(ahp_weight.items()):
            ahp_weight[k] = round(list_scale[i], 6)
        return ahp_weight




if __name__ == '__main__':
    ahp_weight = {
        "robot_detection_score": 1,
        "ip_connection_score": 2,
        "internet_info_score": 2,
        "ip_change_score": 3,
        "device_consistency_score": 4,
        "device_connection_score": 5,
        "personal_device_score": 6,
        "bio_behavior_score": 7
    }
    ahpw = AHPWeight(ahp_weight)
    dict_weight = ahpw.output_weight()
    dict_weight



from sqlalchemy import create_engine


def open_connection(host="192.168.10.106", port="3306", user="root", passwd="root16313302", db="service_report_auth"):
    sql_connect = "mysql+pymysql://{0}:{1}@{2}:{3}/{4}".format(
        user, passwd, host, port, db)
    engine = create_engine(sql_connect, echo=False)
    con = engine.connect()
    return(con)


import pandas as pd
import numpy as np
import sys
import re
from unittest import result
from hitrustai_lab.orm import Orm, get_orm_profile
from hitrustai_lab.orm.Tables.TBMarchant import TbmerchantriskValidationReport, TBmerchantRiskInfo, TBmerchantRiskPredictresult


class UpdateELKAPI:
    def __init__(self, dict_arg):
        self.ELK = dict_arg["obj_ELK"]
        self.table_reason = dict_arg["table_reason"]
        self.abnormal_threshold = dict_arg["abnormal_threshold"]
        self.INDEX_NAME = {
            'info': f'{dict_arg["DATA_SOURCE"]}_{dict_arg["modual_name"]}_info_data',
            'result': f'{dict_arg["DATA_SOURCE"]}_{dict_arg["modual_name"]}_result_data',
            'predict': f'{dict_arg["DATA_SOURCE"]}_{dict_arg["modual_name"]}_predict_data'
        }
        self.associationkey = {
            'info': dict_arg["info_id"],
            'result': dict_arg["result_id"]
            
        }
        self.realTime_mode = dict_arg["realTime_mode"]
        if not dict_arg["realTime_mode"]:
            self.n_rows = dict_arg["n_rows"]
            self.orm = Orm(profile=get_orm_profile(
                host=dict_arg["db_config"]["DB_HOST"],
                db=dict_arg["db_config"]["DB_NAME"],
                user=dict_arg["db_config"]["DB_USERNAME"],
                pwd=dict_arg["db_config"]["DB_PASS"],
                port=dict_arg["db_config"]["DB_PORT"]
            ))

    def get_update_data(self):
        info_start_point = self.ELK.get_start_time(self.INDEX_NAME['info'], date_col='pk_id')
        result_start_point = self.ELK.get_start_time(self.INDEX_NAME['result'], date_col='pk_id')

        if info_start_point is None:
            info_data = self.orm.elk_query_filter(self.dict_arg["orm_predict_info"], self.n_rows)
            result_data = self.orm.elk_query_filter(self.dict_arg["orm_predict_result"], self.n_rows)
        else:
            info_data = self.orm.elk_query_filter(self.dict_arg["orm_predict_info"], self.n_rows, (self.dict_arg["orm_predict_info"].pk_id > info_start_point))
            result_data = self.orm.elk_query_filter(self.dict_arg["orm_predict_result"], self.n_rows, (self.dict_arg["orm_predict_result"].pk_id > result_start_point))

        if info_data.shape[0] == 0 and result_data.shape[0] == 0:
            print(self.dict_arg["modual_name"] + ' Upload Complete ! ')
            return None
        elif info_data.shape[0] == 0 and result_data.shape[0] != 0:
            return {
                'result': result_data
            }
        elif info_data.shape[0] != 0 and result_data.shape[0] == 0:
            return {
                'info': info_data
            }
        else:
            predict_data = self.orm.elk_query_filter(self.dict_arg["orm_predict_result"], self.n_rows, (self.dict_arg["orm_result_id"].in_(info_data['pk_id'].tolist())))
            info_data, predict_data, merge_data = self.preprocess(info_data, predict_data)
            return {
                'info': info_data,
                'result': result_data,
                'predict': merge_data
            }

    def preprocess(self, info_data, result_data):
        merge_data = pd.merge(info_data, result_data, left_on=['pk_id'], right_on=[self.dict_arg["result_id"]], how='inner')
        merge_data = merge_data.reset_index(drop=True)

        # column rename
        merge_data.rename(columns={"pk_id_x": "pk_id", "create_time_y": "create_time","serial_number_x":"serial_number"}, inplace=True)
        merge_data.drop(["pk_id_y", "create_time_x","serial_number_y"], axis=1, inplace=True)
        
        modual_types = list(self.table_reason.keys())
        list_columns_ = []
        list_columns = []
        for modual_name_column in modual_types:
            modual_name = modual_name_column.split("_reason")[0]
            list_columns.append(modual_name)
            merge_data[modual_name_column] = self.table_reason[modual_name_column](merge_data[modual_name_column])
            merge_data[modual_name_column] = merge_data[modual_name_column].astype(str)
            merge_data[modual_name_column + '_'] = np.where(merge_data[modual_name + '_score'] > self.abnormal_threshold, merge_data[modual_name_column], 'nan')
            list_columns_.append(modual_name_column + '_')
        merge_data['abnormal_reason_'] = merge_data[list_columns_].agg(', '.join, axis=1)
        merge_data['abnormal_reason_'] = merge_data['abnormal_reason_'].str.replace(' nan,', '')
        merge_data['abnormal_reason_'] = merge_data['abnormal_reason_'].str.replace('nan, ', '')
        merge_data['abnormal_reason_'] = merge_data['abnormal_reason_'].str.replace(' nan', '')

        merge_data['abnormal_threshold'] = self.abnormal_threshold 
        merge_data['true_label'] = 0
        merge_data["predict_label"] = np.where(merge_data.total_score > self.abnormal_threshold, 1, 0)
        for modual_name in list_columns:
            merge_data[modual_name + "_label"] = np.where(merge_data[modual_name + "_score"] > self.abnormal_threshold, 1, 0)
        # tp fp tn fn
        merge_data['normal_tp'] = np.where(((merge_data["predict_label"] == 0) & (merge_data["predict_label"] == merge_data['true_label'])), 1, 0)
        merge_data['normal_tn'] = np.where(((merge_data["predict_label"] == 1) & (merge_data["predict_label"] == merge_data['true_label'])), 1, 0)
        merge_data['normal_fn'] = np.where(((merge_data["predict_label"] == 1) & (merge_data['true_label'] == 0)), 1, 0)
        merge_data['normal_fp'] = np.where(((merge_data["predict_label"] == 0) & (merge_data["true_label"] == 1)), 1, 0)
        merge_data['abnormal_tp'] = merge_data['normal_tn'].copy()
        merge_data['abnormal_tn'] = merge_data['normal_tp'].copy()
        merge_data['abnormal_fn'] = merge_data['normal_fp'].copy()
        merge_data['abnormal_fp'] = merge_data['normal_fn'].copy()
        merge_data['abnormal_count'] = merge_data['normal_tn'] + merge_data['normal_fn']

        merge_data['year_month'] = pd.to_datetime(merge_data['create_time']).dt.to_period('M').astype(str)
        merge_data['year_month_week'] = pd.to_datetime(merge_data['create_time']).dt.to_period('W').astype(str)
        merge_data['year_date'] = merge_data['create_time'].dt.date
        return info_data, result_data, merge_data

    def get_mappings(self, data):
        dict_type = {}
        for key, value in zip(data.columns, data.dtypes.values):
            if key in ['customer_id', 'profile_id', 'tag', 'model_id', 'training_id', 'data_version', 'mac', 'timestamp', 'udid', 'merchant_id']:
                dict_type[key] = {"type": "text", "fields": {"keyword": {"type": "keyword"}}}
            elif value == 'int64':
                dict_type[key] = {"type": "integer", "fields": {"keyword": {"type": "keyword"}}}
            elif value == 'float64' or 'score' in key:
                dict_type[key] = {"type": "scaled_float", "scaling_factor": 100, "fields": {"keyword": {"type": "keyword"}}}
            elif value == 'O':
                dict_type[key] = {"type": "text", "fields": {"keyword": {"type": "keyword"}}}
            elif 'create_time' in key or 'date' in key:
                dict_type[key] = {"type": "date"}
            else:
                dict_type[key] = {"type": "text", "fields": {"keyword": {"type": "keyword"}}}
        mappings = {
            "mappings": {
                "properties": dict_type
            }
        }
        return mappings

    def elk_insert(self, data_dict=None):
        if self.realTime_mode:
            for key, data in data_dict.items():
                data = data.fillna('undefined')
                mappings = self.get_mappings(data)
                self.ELK.create_index(self.INDEX_NAME[key], mappings)
                result = self.ELK.insert_data(data, self.INDEX_NAME[key])
            del data_dict
            if len(result['fails']) > 0:
                return False
            else:
                return True
        else:
            while True:
                data_dict = self.get_update_data()
                if data_dict != None:
                    for key, data in data_dict.items():
                        data = data.fillna('undefined')
                        mappings = self.get_mappings(data)
                        self.ELK.create_index(self.INDEX_NAME[key], mappings)
                        self.ELK.insert_data(data, self.INDEX_NAME[key])
                    del data_dict
                else:
                    break


if __name__ == '__main__':
    def modify_deviceconsistency_reason(deviceconsistency_reason):
        if deviceconsistency_reason == 'DeviceConsistency-normal':
            deviceconsistency_reason = deviceconsistency_reason.replace('DeviceConsistency-normal', 'Device Consistency - Normal')
        elif deviceconsistency_reason == 'DeviceConsistency-abnormal':
            deviceconsistency_reason = deviceconsistency_reason.replace('DeviceConsistency-abnormal', 'Device Consistency - Abnormal')
        elif deviceconsistency_reason == 'DeviceConsistency-abnormal(ua_model)':
            deviceconsistency_reason = deviceconsistency_reason.replace('DeviceConsistency-abnormal(ua_model)', 'Device Consistency - Abnormal(UA anomaly)')
        return deviceconsistency_reason

    def modify_internetinfo_reason(internetinfo_reason):
        if bool(re.search("InternetInfo-normal", internetinfo_reason)):
            internetinfo_reason = internetinfo_reason.replace('InternetInfo-normal', 'Internet Info - Normal')
        elif bool(re.search("InternetInfo-abnormal", internetinfo_reason)):
            internetinfo_reason = internetinfo_reason.replace('InternetInfo-abnormal', 'Internet Info - Abnormal')
        internetinfo_reason = internetinfo_reason.replace('conflictlanguage', 'language inconsistent in os and browser')
        internetinfo_reason = internetinfo_reason.replace('diffTimezonIPrequestOs', 'timezone inconsistent in ip and os')
        internetinfo_reason = internetinfo_reason.replace('diffTimezonIPrequestOs_withoutTool', 'timezone inconsistent in ip(without proxy) and os')
        internetinfo_reason = internetinfo_reason.replace('diffUseragent', 'request UA and source UA is inconsistent')
        internetinfo_reason = internetinfo_reason.replace('ip_is_tor', 'ip is tor')
        internetinfo_reason = internetinfo_reason.replace('ip_is_proxy', 'ip is proxy')
        internetinfo_reason = internetinfo_reason.replace('ip_cloud_server', 'ip cloud server')
        return internetinfo_reason

    def modify_personaldevice_reason(personaldevice_reason):
        if personaldevice_reason == 'PersonalDevice-normal':
            personaldevice_reason = personaldevice_reason.replace('PersonalDevice-normal', 'Personal Device - Normal')
        elif personaldevice_reason == 'PersonalDevice-normal(NoEnoughHistoryData)':
            personaldevice_reason = personaldevice_reason.replace('PersonalDevice-normal(NoEnoughHistoryData)', 'Insufficient Device Historical Data')
        elif personaldevice_reason == 'PersonalDevice-abnormal(ip_is_proxy)':
            personaldevice_reason = personaldevice_reason.replace('PersonalDevice-abnormal(ip_is_proxy)', 'Personal Device - Abnormal(ip is proxy)')
        elif personaldevice_reason == 'PersonalDevice-abnormal(ip_is_tor)':
            personaldevice_reason = personaldevice_reason.replace('PersonalDevice-abnormal(ip_is_tor)', 'Personal Device - Abnormal(ip is tor)')
        elif personaldevice_reason == 'PersonalDevice-abnormal(ip_is_vpn)':
            personaldevice_reason = personaldevice_reason.replace('PersonalDevice-abnormal(ip_is_vpn)', 'Personal Device - Abnormal(ip is vpn)')
        elif personaldevice_reason == 'PersonalDevice-abnormal(browser_name)':
            personaldevice_reason = personaldevice_reason.replace('PersonalDevice-abnormal(browser_name)', 'Personal Device - Abnormal(browser name infrequently used)')
        elif personaldevice_reason == 'PersonalDevice-abnormal(browser_language)':
            personaldevice_reason = personaldevice_reason.replace('PersonalDevice-abnormal(browser_language)', 'Personal Device - Abnormal(browser language infrequently used)')
        elif personaldevice_reason == 'PersonalDevice-abnormal(os_screen_resolution)':
            personaldevice_reason = personaldevice_reason.replace('PersonalDevice-abnormal(os_screen_resolution)', 'Personal Device - Abnormal(os screen resolution infrequently used)')
        elif personaldevice_reason == 'PersonalDevice-abnormal(os_local_timezone_offset)':
            personaldevice_reason = personaldevice_reason.replace('PersonalDevice-abnormal(os_local_timezone_offset)', 'Personal Device - Abnormal(os local timezone offset infrequently used)')
        elif personaldevice_reason == 'PersonalDevice-abnormal(os_screen_orientation)':
            personaldevice_reason = personaldevice_reason.replace('PersonalDevice-abnormal(os_screen_orientation)', 'Personal Device - Abnormal(os screen orientation infrequently used)')
        elif personaldevice_reason == 'PersonalDevice-abnormal(browser_is_private_mode)':
            personaldevice_reason = personaldevice_reason.replace('PersonalDevice-abnormal(browser_is_private_mode)', 'Personal Device - Abnormal(browser is private mode)')
        return personaldevice_reason

    dict_arg = {
        "obj_ELK": obj_ELK,
        "DATA_SOURCE": "",
        "db_config": dict(),
        "n_rows": "100000",
        "modual_name": "tb_model_fraud_detect",
        "realTime_mode": False,
        "orm_train_report": TbmerchantriskValidationReport,
        "orm_predict_info": TBmerchantRiskInfo,
        "orm_predict_result": TBmerchantRiskPredictresult,
        "table_reason": {
            "device_consistency_reason": np.vectorize(modify_deviceconsistency_reason, otypes=[str]),
            "internet_info_reason": np.vectorize(modify_internetinfo_reason, otypes=[str]),
            "personal_device_reason": np.vectorize(modify_personaldevice_reason, otypes=[str])
        },
        "abnormal_threshold": 0.5,
        "info_id" : "pk_id",
        "result_id" : "mr_info_id"
    }
    UpdateELKAPI(dict_arg)




from datetime import datetime
from sqlalchemy import Column, Integer, Float
from sqlalchemy.dialects.mysql import TIMESTAMP, VARCHAR, DECIMAL
from sqlalchemy.ext.declarative import declarative_base

'''
Scripts to create a table:

# from sqlalchemy import create_engine  

# DBname = 'acqfd_test'
# db_string = f"mysql+pymysql://isrfd:isrfd16313302@35.229.153.69/{DBname}?charset=utf8mb4&binary_prefix=true"
# engine = create_engine(db_string, echo=True)  
# Base = declarative_base() 

# yourtable

# Base.metadata.create_all(engine)  

'''


Base = declarative_base()
# metadata = Base.metadata


class CA_PredictInfo(Base):

    __tablename__ = 'tb_abnormal_transaction_v2_predict_info'

    pk_id = Column(Integer, primary_key=True)
    customer_id = Column(VARCHAR(20))
    training_id = Column(VARCHAR(20))
    profile_id = Column(VARCHAR(20))
    tag = Column(VARCHAR(20))
    model_id = Column(VARCHAR(20))
    data_version = Column(VARCHAR(10))
    mac = Column(VARCHAR(100))
    timestamp = Column(VARCHAR(100))
    sn = Column(VARCHAR(100))
    customer_servertime = Column(TIMESTAMP)
    pan_hash = Column(VARCHAR(128))
    purchase_amount = Column(DECIMAL(16, 5))
    purchase_currency = Column(VARCHAR(45))
    merchant_id = Column(VARCHAR(45))
    udid = Column(VARCHAR(100))
    ip_country = Column(VARCHAR(100))
    os_name = Column(VARCHAR(100))
    browser_language = Column(VARCHAR(100))
    create_time = Column(TIMESTAMP, default=datetime.now)

    def __init__(self, **kwargs):
        for property, value in kwargs.items():
            setattr(self, property, value)


class CA_PredictResult(Base):

    __tablename__ = 'tb_abnormal_transaction_v2_predict_result'

    pk_id = Column(Integer, primary_key=True)
    at_info_id = Column(Integer)
    card_abnormal_score = Column(Float)
    card_abnormal_weight = Column(Float)
    card_abnormal_ahp_score = Column(Float)
    card_abnormal_reason = Column(VARCHAR(300))
    card_abnormal_return_code = Column(VARCHAR(10))
    total_score = Column(Float)
    policy_score = Column(Float)
    create_time = Column(TIMESTAMP, default=datetime.now)

    def __init__(self, **kwargs):
        for property, value in kwargs.items():
            setattr(self, property, value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time:2020.09.12
# @author:xhrg
# @email:634789257@qq.com
import logging
import os
import urllib2
from urllib import urlencode


def http_request(url, timeout, headers={}):
    try:
        request = urllib2.Request(url, headers=headers)
        res = urllib2.urlopen(request, timeout=timeout)
        body = res.read().decode("utf-8")
        return res.code, body
    except urllib2.HTTPError as e:
        if e.code == 304:
            logging.getLogger(__name__).warning("http_request error,code is 304, maybe you should check secret")
            return 304, None
        logging.getLogger(__name__).warning("http_request error,code is %d, msg is %s", e.code, e.msg)
        raise e


def url_encode(params):
    return urlencode(params)


def makedirs_wrapper(path):
    try:
        os.makedirs(path, mode=0755)
    except OSError:
        pass



import numpy as np
import pandas as pd 
from sklearn.metrics import confusion_matrix
from sklearn.linear_model import LinearRegression
from sklearn.model_selection import train_test_split
from hitrustai_lab.algorithm.ahp import AHPWeight

'''
auth_lable過濾auth code哪寫為異常
get_best_score : 算出整份auth預測與真實答案的最佳界線(多少policy_score以下為異常)
'''


def auth_lable(df_auth_columns, list_auth_code):
    return np.where((df_auth_columns.isin(list_auth_code)), 0, 1)


# get_best_score(df_new,"policy_score_fd","true_label")
def get_best_score(df_inpute_data, policy_score_col_name, true_label_col_name, predict_label_col_name):
    arr_list = np.arange(-0.9, 1, 0.05)
    list_a = []
    list_b = []
    list_score = []
    for i in arr_list:
        df_inpute_data[predict_label_col_name] = np.where((df_inpute_data[policy_score_col_name] < i), 0, 1)
        matrix = confusion_matrix(df_inpute_data[true_label_col_name], df_inpute_data[predict_label_col_name])
        a = matrix[0][1]
        b = matrix[1][0]
        list_a.append(a)
        list_b.append(b)
        list_score.append(i)
    grades = {
        "ideal": list_a,
        "srcond_ideal": list_b,
        "score": list_score
    }
    df = pd.DataFrame(grades)
    exam_x = df['ideal']
    exam_y = df['srcond_ideal']

    x_train, x_test, y_train, y_test = train_test_split(
        exam_x, exam_y, train_size=.9)

    x_train = x_train.values.reshape(-1, 1)
    y_train = y_train.values.reshape(-1, 1)
    x_test = x_test.values.reshape(-1, 1)
    y_test = y_test.values.reshape(-1, 1)
    
    x1 = exam_x.values.reshape(-1, 1)
    y1 = exam_y.values.reshape(-1, 1)
    model = LinearRegression()

    model.fit(x1, y1)
    y_train_pred = model.predict(x1)
    s = np.absolute(y_train_pred - y1).astype(int)
    best_score = None
    min_ideal = None
    for i in np.where(s == min(s))[0]:
        m = df.loc[df.index == i]["score"].values[0]
        ideal = df.loc[df.index == i]["ideal"].values[0]

        if min_ideal is None:
            min_ideal = ideal
            best_score = round(m, 2)
        else:
            if min_ideal > ideal:
                min_ideal = ideal
                best_score = round(m, 2)
    return best_score


# coding: utf-8
from sqlalchemy import Column, text, Integer, FLOAT
from sqlalchemy.dialects.mysql import LONGTEXT, VARCHAR, TIMESTAMP
from sqlalchemy.ext.declarative import declarative_base

# db_string = "mysql+pymysql://"+'acqfd'+":"+'acqfd16313302'+"@"+'192.168.10.201'+"/"+'acqfd_test'+"?charset=utf8mb4&binary_prefix=true"
# engine = create_engine(db_string, echo=True)
# Base = declarative_base()

Base = declarative_base()
metadata = Base.metadata


class Deviceinfo(Base):
    __tablename__ = 'deviceinfo'

    pk_id = Column(Integer, primary_key=True, autoincrement=True, unique=True)
    serial_number = Column(VARCHAR(100), nullable=False)
    client_account = Column(VARCHAR(20), nullable=False)
    udid = Column(VARCHAR(100), nullable=False)
    hardware_device_type = Column(VARCHAR(100))
    hardware_device_model = Column(VARCHAR(100))
    hardware_device_brand = Column(VARCHAR(100))
    hardware_cpu_architecture = Column(VARCHAR(100))
    hardware_cpu_cores = Column(VARCHAR(100))
    hardware_memory_size = Column(VARCHAR(100))
    hardware_gpu_name = Column(VARCHAR(200))
    hardware_is_battery_charging = Column(VARCHAR(10))
    hardware_battery_level = Column(VARCHAR(10))
    hardware_audio_fp = Column(VARCHAR(100))
    hardware_pointing_method = Column(VARCHAR(100))
    hardware_max_touch_points = Column(VARCHAR(10))
    # hardware_connection = Column(VARCHAR(10))
    os_name = Column(VARCHAR(100))
    os_version = Column(VARCHAR(100))
    os_platform = Column(VARCHAR(100))
    os_local_datetime = Column(VARCHAR(100))
    os_local_timezone = Column(VARCHAR(100))
    os_local_timezone_offset = Column(VARCHAR(100))
    os_screen_orientation = Column(VARCHAR(100))
    os_screen_resolution_width = Column(VARCHAR(10))
    os_screen_resolution_height = Column(VARCHAR(10))
    os_screen_pixel_ratio = Column(VARCHAR(30))
    os_screen_color_depth = Column(VARCHAR(20))
    ip_source = Column(VARCHAR(20))
    ip_foundation = Column(VARCHAR(20))
    ip_intranet = Column(VARCHAR(100))
    ip_src_latitude = Column(VARCHAR(15))
    ip_src_longitude = Column(VARCHAR(15))
    ip_src_timezone = Column(VARCHAR(100))
    ip_src_country = Column(VARCHAR(100))
    ip_request = Column(VARCHAR(20))
    ip_latitude = Column(VARCHAR(15))
    ip_longitude = Column(VARCHAR(15))
    ip_timezone = Column(VARCHAR(100))
    ip_country = Column(VARCHAR(100))
    # ip_region = Column(VARCHAR(100))
    # ip_city = Column(VARCHAR(100))
    # ip_suburb = Column(VARCHAR(100))
    # ip_city_district = Column(VARCHAR(100))
    # ip_village = Column(VARCHAR(100))
    # ip_road = Column(VARCHAR(100))
    ip_isp = Column(VARCHAR(100))
    ip_hostname = Column(VARCHAR(100))
    ip_is_proxy = Column(VARCHAR(10))
    ip_is_tor = Column(VARCHAR(10))
    ip_is_vpn = Column(VARCHAR(10))
    ip_cloud_server = Column(VARCHAR(100))
    # ip_dnsbl = Column(VARCHAR(10))
    # ip_surbl = Column(VARCHAR(10))
    browser_name = Column(VARCHAR(100))
    browser_version = Column(VARCHAR(100))
    browser_engine = Column(VARCHAR(100))
    browser_resolution_width = Column(VARCHAR(10))
    browser_resolution_height = Column(VARCHAR(10))
    browser_url = Column(VARCHAR(500))
    browser_useragent = Column(VARCHAR(500))
    browser_req_useragent = Column(VARCHAR(500))
    browser_plugins = Column(LONGTEXT)
    browser_mimetypes = Column(VARCHAR(500))
    browser_language = Column(VARCHAR(100))
    browser_language_fp = Column(VARCHAR(100))
    browser_is_support_canvas = Column(VARCHAR(10))
    browser_canvas_fp = Column(VARCHAR(100))
    browser_number_of_available_fonts = Column(VARCHAR(100))
    browser_available_fonts = Column(LONGTEXT)
    browser_fonts_fp = Column(VARCHAR(100))
    browser_is_support_webgl1 = Column(VARCHAR(10))
    browser_is_support_webgl2 = Column(VARCHAR(10))
    browser_webgl_fp = Column(VARCHAR(100))
    # browser_is_tracking_protection = Column(VARCHAR(10))
    browser_is_support_cookies = Column(VARCHAR(10))
    # browser_is_support_flash = Column(VARCHAR(10))
    # browser_is_support_activex = Column(VARCHAR(10))
    # browser_is_support_javaapplet = Column(VARCHAR(10))
    # browser_is_support_audio_context = Column(VARCHAR(10))
    browser_is_support_webrtc = Column(VARCHAR(10))
    browser_is_support_websocket = Column(VARCHAR(10))
    browser_is_support_indexeddb = Column(VARCHAR(10))
    browser_is_support_local_storage = Column(VARCHAR(10))
    browser_is_support_session_storage = Column(VARCHAR(10))
    browser_is_support_open_database = Column(VARCHAR(10))
    browser_http_version = Column(VARCHAR(20))
    # browser_tls_protocol = Column(VARCHAR(20))
    # browser_tls_cipher = Column(VARCHAR(100))
    # browser_accepted_content_types = Column(VARCHAR(100))
    browser_accepted_content_encodings = Column(VARCHAR(100))
    browser_is_private_mode = Column(VARCHAR(10))
    browser_is_use_webdriver = Column(VARCHAR(10))
    browser_is_brave = Column(VARCHAR(10))
    browser_is_tor = Column(VARCHAR(10))
    # browser_is_use_adblock = Column(VARCHAR(10))
    # browser_pageviews = Column(VARCHAR(10))
    # browser_account_login = Column(VARCHAR(100))
    create_time = Column(TIMESTAMP(fsp=6), nullable=False,
                         server_default=text("CURRENT_TIMESTAMP(6)"))


class UDIDHistory(Base):
    __tablename__ = 'udid_history'

    pk_id = Column(Integer, primary_key=True, autoincrement=True, unique=True)
    indexeddb_version = Column(VARCHAR(45))
    indexeddb_pointer = Column(VARCHAR(100))
    internet_ip = Column(VARCHAR(20))
    intranet_ip = Column(VARCHAR(20))
    udid_pointer_id = Column(Integer)
    deviceinfo_id = Column(Integer)
    udid_path = Column(VARCHAR(5))
    create_time = Column(TIMESTAMP(fsp=6), nullable=False,
                         server_default=text("CURRENT_TIMESTAMP(6)"))


class RobotInfo(Base):
    __tablename__ = 'robot_info'

    pk_id = Column(Integer, primary_key=True, autoincrement=True)
    serial_number = Column(VARCHAR(100))
    client_account = Column(VARCHAR(100))
    is_robot = Column(VARCHAR(10))
    robot_val = Column(FLOAT)
    # robot_score = Column(FLOAT)
    # bio_score = Column(FLOAT)
    device_type = Column(VARCHAR(100))
    platform = Column(VARCHAR(100))
    vendor = Column(VARCHAR(100))
    renderer = Column(VARCHAR(100))
    os_name = Column(VARCHAR(100))
    is_emulator = Column(VARCHAR(10))
    is_proxy = Column(VARCHAR(10))
    is_tor = Column(VARCHAR(10))
    is_vpn = Column(VARCHAR(10))
    is_cloud = Column(VARCHAR(10))
    browser_name = Column(VARCHAR(100))
    useragent = Column(VARCHAR(200))
    headless = Column(VARCHAR(10))
    webdriver = Column(VARCHAR(10))
    is_private_mode = Column(VARCHAR(10))
    eval = Column(Integer)
    mouse_enter = Column(VARCHAR(10))
    mouse_click = Column(VARCHAR(10))
    touch_start = Column(VARCHAR(10))
    tab = Column(VARCHAR(10))
    paste = Column(VARCHAR(10))
    focus_on_abnormal = Column(VARCHAR(10))
    keyin_abnormal = Column(VARCHAR(10))
    keydown_avg_time = Column(FLOAT)
    is_change_useragent = Column(VARCHAR(10))
    is_change_plugins = Column(VARCHAR(10))
    is_change_languages = Column(VARCHAR(10))
    is_change_webdriver = Column(VARCHAR(10))
    is_change_eval = Column(VARCHAR(10))
    is_change_platform = Column(VARCHAR(10))
    is_short_page_time = Column(VARCHAR(10))
    page_time = Column(FLOAT)
    # mouse_log = Column(LONGTEXT)
    # events_log = Column(LONGTEXT)
    create_time = Column(TIMESTAMP(fsp=6), nullable=False,
                         server_default=text("CURRENT_TIMESTAMP(6)"))


class RobotInfoV2(Base):
    __tablename__ = 'robot_info_v2'

    pk_id = Column(Integer, primary_key=True, autoincrement=True)
    serial_number = Column(VARCHAR(100))
    client_account = Column(VARCHAR(100))
    # is_robot = Column(VARCHAR(10))
    # robot_val = Column(FLOAT)
    # robot_score = Column(FLOAT)
    # bio_score = Column(FLOAT)
    # device_type = Column(VARCHAR(100))
    platform = Column(VARCHAR(100))
    vendor = Column(VARCHAR(100))
    renderer = Column(VARCHAR(100))
    # os_name = Column(VARCHAR(100))
    # is_emulator = Column(VARCHAR(10))
    # is_proxy = Column(VARCHAR(10))
    # is_tor = Column(VARCHAR(10))
    # is_vpn = Column(VARCHAR(10))
    # is_cloud = Column(VARCHAR(10))
    # browser_name = Column(VARCHAR(100))
    useragent = Column(VARCHAR(200))
    # headless = Column(VARCHAR(10))
    webdriver = Column(VARCHAR(10))
    is_private_mode = Column(VARCHAR(10))
    eval = Column(Integer)
    mouse_enter = Column(VARCHAR(10))
    mouse_click = Column(VARCHAR(10))
    touch_start = Column(VARCHAR(10))
    tab = Column(VARCHAR(10))
    paste = Column(VARCHAR(10))
    # focus_on_abnormal = Column(VARCHAR(10))
    # keyin_abnormal = Column(VARCHAR(10))
    # keydown_avg_time = Column(FLOAT)
    is_change_useragent = Column(VARCHAR(10))
    is_change_plugins = Column(VARCHAR(10))
    is_change_languages = Column(VARCHAR(10))
    is_change_webdriver = Column(VARCHAR(10))
    is_change_eval = Column(VARCHAR(10))
    is_change_platform = Column(VARCHAR(10))
    # is_short_page_time = Column(VARCHAR(10))
    page_time = Column(FLOAT)
    # mouse_log = Column(LONGTEXT)
    # events_log = Column(LONGTEXT)
    resolution_width = Column(VARCHAR(10))
    resolution_height = Column(VARCHAR(10))
    pixel_ratio = Column(VARCHAR(10))
    source_ip = Column(VARCHAR(20))
    create_time = Column(TIMESTAMP(fsp=6), nullable=False,
                         server_default=text("CURRENT_TIMESTAMP(6)"))



import os
import re
import json
import time
import datetime
import pandas as pd
import numpy as np
from decouple import AutoConfig
from sqlalchemy import Column, text, Integer
from sqlalchemy.dialects.mysql import TIMESTAMP

from hitrustai_lab.model_train.core.utils import decrypt_passwd
from hitrustai_lab.apollo.apollo_client import ApolloClient
from hitrustai_lab.apollo.util import check_apollo_change, get_apollo_value

from hitrustai_lab.matrix.model_performance import ModelPerfornance
from hitrustai_lab.orm import Orm, get_orm_profile
from hitrustai_lab.orm.Tables.ModelPerformance import create_model_performance, Base_Model


dict_init_arg = {
    "list_y_test": list,
    "list_y_score": np.array,
    "customer_id_lst": str,
    "training_id_lst": str,
    "model_id_lst": str,
    "profile_id_lst": str,
    "tag_lst": str,
    "training_start_time_lst": str,
    "total_training_time_lst": int,
    "training_start_time_lst": datetime,
    "training_end_time_lst": datetime,
    "number_of_dump_data": int,
    "number_of_training_data": int,
    "number_of_positive_samples_in_training_data": int,
    "number_of_negative_samples_in_training_data": int,
    "number_of_validation_data": int,
    "true_label_column_lst": str,
    "number_of_positive_samples_in_validation_data": int,
    "number_of_negative_samples_in_validation_data": int,
    "return_code": str,
    "reason": str
}


class TrainModelToSQl:
    def __init__(self, host="192.168.10.102", port="3305", user="root", passwd="root16313302", db="diia_test", table_name="int") -> None:
        self.table_name = table_name
        self.orm_profile = get_orm_profile(host=host, port=port, db=db, user=user, pwd=passwd)
        self.orm = Orm(profile=self.orm_profile)

    def performance(self, dict_init_arg: dict):
        mp = ModelPerfornance(score_type='policy_score')
        try:
            result = mp.performance_output(dict_init_arg["list_y_test"], dict_init_arg["list_y_score"])
            result = {
                'customer_id': dict_init_arg["customer_id_lst"],
                'training_id': dict_init_arg["training_id_lst"],
                'model_id': dict_init_arg["model_id_lst"],
                'profile_id': dict_init_arg["profile_id_lst"],
                'tag': dict_init_arg["tag_lst"],
                'model_name': dict_init_arg["model_id_lst"] + "-" + "".join(re.findall(r'\d+', str(datetime.datetime.now()))),
                'training_start_time': dict_init_arg["training_start_time_lst"],
                'training_end_time': dict_init_arg["training_end_time_lst"],
                'total_training_time': dict_init_arg["total_training_time_lst"],
                'training_data_start_date': dict_init_arg["training_data_start_date"],
                'training_data_end_date': dict_init_arg["training_data_end_date"],
                'number_of_dump_data': dict_init_arg["number_of_dump_data"],
                'number_of_training_data': dict_init_arg["number_of_training_data"],
                'number_of_positive_samples_in_training_data': dict_init_arg["number_of_positive_samples_in_training_data"],
                'number_of_negative_samples_in_training_data': dict_init_arg["number_of_negative_samples_in_training_data"],
                'number_of_validation_data': dict_init_arg["number_of_validation_data"],
                'true_label_column': dict_init_arg["true_label_column_lst"],
                'number_of_positive_samples_in_validation_data': dict_init_arg["number_of_positive_samples_in_validation_data"],
                'number_of_negative_samples_in_validation_data': dict_init_arg["number_of_negative_samples_in_validation_data"],
                'threshold': [result['threshold_lst']],
                'tp': [result['tp_lst']],
                'fp': [result['fp_lst']],
                'tn': [result['tn_lst']],
                'fn': [result['fn_lst']],
                'accuracy': [result['accuracy_lst']],
                'ppv': [result['precision_lst']],
                'recall': [result['recall_lst']],
                'f1_score': [result['f1_score_lst']],
                'fnr': [result['fnr_lst']],
                'fpr': [result['fpr_lst']],
                'npv': [result['npv_lst']],
                'fdr': [result['fdr_lst']],
                'for_': [result['for_lst']],
                'tnr': [result['tnr_lst']],
                'auc': result['auc_lst'],
                "return_code": dict_init_arg['return_code'],
                "reason": dict_init_arg['reason']
            }
        except Exception:
            result = {
                'customer_id': dict_init_arg["customer_id_lst"],
                'training_id': dict_init_arg["training_id_lst"],
                'model_id': dict_init_arg["model_id_lst"],
                'profile_id': dict_init_arg["profile_id_lst"],
                'tag': dict_init_arg["tag_lst"],
                'model_name': dict_init_arg["model_id_lst"] + "-" + "".join(re.findall(r'\d+', str(datetime.datetime.now()))),
                'training_start_time': "2000-01-01 00:00:00",
                'training_end_time': "2000-01-01 00:00:00",
                'total_training_time': -1,
                'training_data_start_date': "2000-01-01 00:00:00",
                'training_data_end_date': "2000-01-01 00:00:00",
                'number_of_dump_data': -1,
                'number_of_training_data': -1,
                'number_of_positive_samples_in_training_data': -1,
                'number_of_negative_samples_in_training_data': -1,
                'number_of_validation_data': -1,
                'true_label_column': dict_init_arg["true_label_column_lst"],
                'number_of_positive_samples_in_validation_data': -1,
                'number_of_negative_samples_in_validation_data': -1,
                'threshold': [[]],
                'tp': [[]],
                'fp': [[]],
                'tn': [[]],
                'fn': [[]],
                'accuracy': [[]],
                'ppv': [[]],
                'recall': [[]],
                'f1_score': [[]],
                'fnr': [[]],
                'fpr': [[]],
                'npv': [[]],
                'fdr': [[]],
                'for_': [[]],
                'tnr': [[]],
                'auc': [[]],
                "return_code": dict_init_arg['return_code'],
                "reason": dict_init_arg['reason']
            }   
        return result

    def dict_to_dataframe(self, dict_init_arg: dict, dict_add_column=None):
        self.need_column = self.performance(dict_init_arg)
        if dict_add_column is not None:
            for key in dict_add_column:
                self.need_column[key] = dict_init_arg[key]
        
        df = pd.DataFrame(data=self.need_column)
        df['total_training_time'] = df.total_training_time
        df['training_data_start_date'] = df.training_data_start_date
        df['training_data_end_date'] = df.training_data_end_date

        for col in [
            'threshold', 'tp', 'fp', 'tn', 'fn', 'accuracy', 'ppv', 'recall', 'f1_score',
            'fnr', 'fpr', 'npv', 'fdr', 'for_', 'tnr'
        ]:
            df[col] = df[col].apply(lambda x: json.dumps(x))

        return df

    def insert_db(self, data: dict, dict_add_column=None):
        # if not data:
        #     return
        base_table = create_model_performance(self.table_name)
        if dict_add_column is not None:
            for key in dict_add_column:
                base_table.append_column(dict_add_column[key])
        base_table.append_column(Column("create_time", TIMESTAMP(fsp=6), nullable=False, server_default=text("CURRENT_TIMESTAMP(6)")))
        
        class User(Base_Model):
            __table__ = base_table
        self.orm.create_table(Base_Model, User)
        
        data = self.dict_to_dataframe(data, dict_add_column)
        
        self.orm.data_to_DB(data, User)


class GetConfArg:
    def read_train_conf_env(self):
        config = AutoConfig(search_path=os.getcwd() + "/env")
        # check docker environment exist
        CUSTOMER_ID = os.environ.get('CUSTOMER_ID')
        MODEL_ID = os.environ.get('MODEL_ID')
        TRAINING_ID = os.environ.get('TRAINING_ID')
        KAFKA_TOPIC = os.environ.get('KAFKA_TOPIC')

        # check service .env file exist
        kafka_node = config('KAFKA_N', default='0')
        bootstrap_servers = []
        for i in range(int(kafka_node)):
            host = config('KAFKA_HOST_' + str(i + 1), default='')
            port = config('KAFKA_PORT_' + str(i + 1), default='')
            if host == "" or port == "":
                self.init_logger.error('Error: Invalid variable in env. Expected keys [KAFKA_HOST, KAFKA_PORT]')
                os.kill(0, 4)
            bootstrap_servers.append(host + ":" + port)

        self.appollo_info = {
            "servers": bootstrap_servers,
            "customer_id": CUSTOMER_ID,
            "model_id": MODEL_ID,
            "training_id": TRAINING_ID,
            "topic": KAFKA_TOPIC,
            "BATCH_SIZE": int(config('Chunksize', default='100000')),
            "dataset_path": config('SOURCE_PATH_DATASET', default=''),
            "kg_path": config('SOURCE_PATH_KNOWLEDGE', default=''),
            "lib_path": config('SOURCE_PATH_LIB', default=''),
            "KAFKA_USE": config('KAFKA_USE', default=''),
        }

        db_pass = config('DB_PASS', default='')
        password = decrypt_passwd(self.passwd_so_name, db_pass)
        self.SQLALCHEMY_DATABASE_URI = '{}://{}:{}@{}:{}/{}'.format(
            config('DB_ENGINE', default='mysql+pymysql'),
            config('DB_USERNAME', default='test'),
            password,
            config('DB_HOST', default='127.0.0.1'),
            config('DB_PORT', default=3306),
            config('DB_NAME', default='testdb')
        )

    def read_train_conf_apollo(self):
        config = AutoConfig(search_path=os.getcwd() + "/env")
        try:
            APOLLO_URL = config('APOLLO_URL')
            APOLLO_APPID = config('APOLLO_APPID')
            APOLLO_CLUSTER = config('APOLLO_CLUSTER')
            APOLLO_SECRET = config('APOLLO_SECRET')
            APOLLO_NAMESPACE_INF = config('APOLLO_NAMESPACE_INF')
            APOLLO_NAMESPACE_MODAL = config('APOLLO_NAMESPACE_MODAL')
        except Exception:
            os.kill(0, 4)
        apollo_client = ApolloClient(
            app_id=APOLLO_APPID,
            cluster=APOLLO_CLUSTER,
            config_url=APOLLO_URL,
            secret=APOLLO_SECRET,
            change_listener=check_apollo_change)

        # check service .env file exist
        KAFKA_N = get_apollo_value(apollo_client, "KAFKA_NODE", APOLLO_NAMESPACE_INF)
        KAFKA_N = KAFKA_N.split(",")
        # kafka_node = len(KAFKA_N)
        bootstrap_servers = []
        for ip in KAFKA_N:
            host, port = ip.split(":")
            if host == "" or port == "":
                self.init_logger.error('Error: Invalid variable in env. Expected keys [KAFKA_HOST, KAFKA_PORT]')
                os.kill(0, 4)
            bootstrap_servers.append(host + ":" + port)
        db_pass = get_apollo_value(apollo_client, "DB_PASS", APOLLO_NAMESPACE_INF)

        password = decrypt_passwd(self.passwd_so_name, db_pass)
        self.SQLALCHEMY_DATABASE_URI = '{}://{}:{}@{}:{}/{}'.format(
            get_apollo_value(apollo_client, "DB_ENGINE", APOLLO_NAMESPACE_INF),
            get_apollo_value(apollo_client, "DB_USERNAME", APOLLO_NAMESPACE_INF),
            password,
            get_apollo_value(apollo_client, "DB_HOST", APOLLO_NAMESPACE_INF),
            get_apollo_value(apollo_client, "DB_PORT", APOLLO_NAMESPACE_INF),
            get_apollo_value(apollo_client, "DB_NAME", APOLLO_NAMESPACE_INF)
        )
        self.appollo_info = {
            "servers": bootstrap_servers,
            "customer_id": get_apollo_value(apollo_client, "CUSTOMER_ID", APOLLO_NAMESPACE_MODAL),
            "model_id": get_apollo_value(apollo_client, "MODEL_ID", APOLLO_NAMESPACE_MODAL),
            "training_id": get_apollo_value(apollo_client, "TRAINING_ID", APOLLO_NAMESPACE_MODAL),
            "topic": get_apollo_value(apollo_client, "TRAINING_ID", APOLLO_NAMESPACE_MODAL),
            "BATCH_SIZE": int(get_apollo_value(apollo_client, "BATCH_SIZE", APOLLO_NAMESPACE_MODAL)),

            "dataset_path": get_apollo_value(apollo_client, "SOURCE_PATH_DATASET", APOLLO_NAMESPACE_MODAL),
            "kg_path": get_apollo_value(apollo_client, "SOURCE_PATH_KNOWLEDGE", APOLLO_NAMESPACE_MODAL),
            "lib_path": get_apollo_value(apollo_client, "SOURCE_PATH_LIB", APOLLO_NAMESPACE_MODAL),
            "KAFKA_USE": get_apollo_value(apollo_client, "KAFKA_USE", APOLLO_NAMESPACE_MODAL),
            "DB_ENGINE": get_apollo_value(apollo_client, "DB_ENGINE", APOLLO_NAMESPACE_INF),
            "DB_USERNAME": get_apollo_value(apollo_client, "DB_USERNAME", APOLLO_NAMESPACE_INF),
            "DB_PASS": password,
            "DB_HOST": get_apollo_value(apollo_client, "DB_HOST", APOLLO_NAMESPACE_INF),
            "DB_PORT": get_apollo_value(apollo_client, "DB_PORT", APOLLO_NAMESPACE_INF),
            "DB_NAME": get_apollo_value(apollo_client, "DB_NAME", APOLLO_NAMESPACE_INF)
        }


class HitrustaiTrainTemple(GetConfArg):
    def __init__(self, dict_model, init_logger, passwd_so_name="./data/passwd.so", model_name="fraud detect"):
        self.dict_model = dict_model
        self.init_logger = init_logger
        self.passwd_so_name = passwd_so_name
        self.model_name = model_name
        self.dict_init_arg = None
        self.template_output_txt = '''
=================================================
INFO_MODEL_NAME: %s
INFO_TRAINING_ID : %s
INFO_CUSTOMER_ID: %s
INFO_MODEL_ID: %s
DATA_START_DATE: %s
DATA_END_DATE: %s
DATA_TOTAL_ROW: %s
TRAINING_DATE: %s
TRAINING_TIME: %s
TRAINING_RETURN_CODE : %s
TRAINING_REASON : %s
        '''

    def error_arg_dict(self):
        dict_init_arg = {
            "list_y_test": [],
            # "list_y_score": np.array(self.df["total_score_fd7"]),
            "list_y_score": [],
            "customer_id_lst": self.appollo_info["customer_id"],
            "training_id_lst": self.appollo_info["training_id"],
            "model_id_lst": self.appollo_info["model_id"],
            "profile_id_lst": self.appollo_info["customer_id"],
            "tag_lst": self.appollo_info["customer_id"],
            "training_start_time_lst": self.training_start_time_lst,
            "total_training_time_lst": self.total_training_time_lst,

            "training_data_start_date": self.DATA_START_DATE,
            "training_data_end_date": self.DATA_END_DARE,
            "training_end_time_lst": self.training_end_time_lst,
            "number_of_dump_data": -1,
            "number_of_training_data": -1,
            "number_of_positive_samples_in_training_data": -1,
            "number_of_negative_samples_in_training_data": -1,
            "number_of_validation_data": -1,
            "true_label_column_lst": self.dict_model.true_lable_name,
            "number_of_positive_samples_in_validation_data": -1,
            "number_of_negative_samples_in_validation_data": -1,
            "return_code": self.dict_report["return_code"],
            "reason": self.dict_report["reason"]
        }
        return dict_init_arg

    def input_arg_dict(self):
        """
        dict_init_arg = {
            "list_y_test": list,
            "list_y_score": np.array,
            "customer_id_lst": str,
            "training_id_lst": str,
            "model_id_lst": str,
            "profile_id_lst": str,
            "tag_lst": str,
            "training_start_time_lst": str,
            "total_training_time_lst": int,
            "training_start_time_lst": datetime,
            "training_end_time_lst": datetime,
            "number_of_dump_data": int,
            "number_of_training_data": int,
            "number_of_positive_samples_in_training_data": int,
            "number_of_negative_samples_in_training_data": int,
            "number_of_validation_data": int,
            "true_label_column_lst": str,
            "number_of_positive_samples_in_validation_data": int,
            "number_of_negative_samples_in_validation_data": int,
            "return_code": str,
            "reason": str
        }
        """
        
        dict_init_arg = {
            "list_y_test": list(self.dict_model.df[self.dict_model.true_lable_name]),
            # "list_y_score": np.array(self.df["total_score_fd7"]),
            "list_y_score": np.array,
            "customer_id_lst": self.appollo_info["customer_id"],
            "training_id_lst": self.appollo_info["training_id"],
            "model_id_lst": self.appollo_info["model_id"],
            "profile_id_lst": self.appollo_info["customer_id"],
            "tag_lst": self.appollo_info["customer_id"],
            "training_start_time_lst": self.training_start_time_lst,
            "total_training_time_lst": self.total_training_time_lst,

            "training_data_start_date": self.DATA_START_DATE,
            "training_data_end_date": self.DATA_END_DARE,
            "training_end_time_lst": self.training_end_time_lst,
            "number_of_dump_data": self.dict_model.DATA_TOTAL_ROW,
            "number_of_training_data": self.dict_model.DATA_TOTAL_ROW_TRAIN,
            "number_of_positive_samples_in_training_data": self.dict_model.VAL_POSITION_NUMBER,
            "number_of_negative_samples_in_training_data": self.dict_model.VAL_POSITION_NEGATIVE,
            "number_of_validation_data": self.dict_model.DATA_TOTAL_ROW_VAL,
            "true_label_column_lst": self.dict_model.true_lable_name,
            "number_of_positive_samples_in_validation_data": self.dict_model.VAL_POSITION_NUMBER_VAL,
            "number_of_negative_samples_in_validation_data": self.dict_model.VAL_POSITION_NEGATIVE_VAL,
            "return_code": self.dict_report["return_code"],
            "reason": self.dict_report["reason"]
        }
        return dict_init_arg

    def train(self):
        try:
            config = AutoConfig(search_path=os.getcwd() + "/env")
            if config('ENV_METHOD') == "env":
                self.read_train_conf_env()
            else:
                self.read_train_conf_apollo()
            t1 = time.time()
            self.training_start_time_lst = datetime.datetime.now()
            self.dict_report = self.dict_model.train()
            self.total_training_time_lst = time.time() - t1
            self.training_end_time_lst = datetime.datetime.now()
            self.DATA_START_DATE = self.dict_model.DATA_START_DATE
            self.DATA_END_DARE = self.dict_model.DATA_END_DARE
            self.DATA_TOTAL_ROW = self.dict_model.DATA_TOTAL_ROW

            self.dict_init_arg = self.input_arg_dict()

        except Exception as e:
            self.training_start_time_lst = "2000-01-01 00:00:00"
            self.total_training_time_lst = -1
            self.training_end_time_lst = "2000-01-01 00:00:00"
            self.DATA_START_DATE = "2000-01-01 00:00:00"
            self.DATA_END_DARE = "2000-01-01 00:00:00"
            self.DATA_TOTAL_ROW = -1
            
            self.dict_init_arg = self.error_arg_dict()
            self.dict_init_arg["reason"] = e.args[0]
            if self.dict_report["return_code"] == "4001":
                self.dict_init_arg["return_code"] = "9914"

    def insert_db(self, table_name, dict_add_column=None):
        tmts = TrainModelToSQl(
            host=self.appollo_info["DB_HOST"],
            port=self.appollo_info["DB_PORT"],
            user=self.appollo_info["DB_USERNAME"],
            passwd=self.appollo_info["DB_PASS"],
            db=self.appollo_info["DB_NAME"],
            table_name=table_name
        )
        tmts.insert_db(data=self.dict_init_arg, dict_add_column=dict_add_column)

    def output_txt(self, model_name="Fraud Detect", path='./data/lib/output.txt'):
        output_txt = self.template_output_txt % (
            model_name,
            self.dict_init_arg["training_id_lst"],
            self.dict_init_arg["customer_id_lst"],
            self.dict_init_arg["model_id_lst"],
            self.dict_init_arg["training_data_start_date"],
            self.dict_init_arg["training_data_end_date"],
            self.dict_init_arg["number_of_dump_data"],
            self.dict_init_arg["training_start_time_lst"],
            self.dict_init_arg["total_training_time_lst"],
            self.dict_init_arg["return_code"],
            self.dict_init_arg["reason"],
        )
        with open(path, 'a') as f:
            f.write(output_txt + "\n")
        

if __name__ == '__main__':
    from hitrustai_lab.model_train.ai_module_train import TrainModelToSQl
    dict_add_column = {
        "add_column1": Column("add_column1", Integer, primary_key=True),
        "add_column2": Column("add_column2", Integer, primary_key=True)
    }

    tmts = TrainModelToSQl(
        host="192.168.10.203",
        port="3305",
        user="diia",
        passwd="diia16313302",
        db="service_report",
        table_name="test111111"
    )

    tmts.insert_db(data=dict_init_arg, dict_add_column=dict_add_column)



import logging
import logging.config


class LogHandler:
    def __init__(self, log_level='INFO', service='CardTesting'):
        self.log_level = log_level
        self.service = service
        self.set_logging()

    def set_logging(self):
        logging.config.dictConfig(config=self._get_logging_config())

    def _get_logging_config(self):
        return {
            "version": 1,
            "disable_existing_loggers": True,
            "formatters": {
                "normal": {  # the name of formatter
                    "format": "%(asctime)s|%(levelname)s|{}|%(name)s|[%(pathname)s, %(lineno)s] : %(message)s ".format(self.service)
                }
            },
            "handlers": {
                "console1": {  # the name of handler
                    # emit to sys.stderr(default)
                    "class": "logging.StreamHandler",
                    "formatter": "normal",  # use the above "normal" formatter
                }
            },
            "loggers": {
                "INIT": {  # the name of logger
                    # use the above "console1" and "console2" handler
                    "handlers": ["console1"],
                    "level": self.log_level  # logging level
                },
                "SERVICE": {
                    "handlers": ["console1"],
                    "level": self.log_level
                },
                "KERNEL": {
                    "handlers": ["console1"],
                    "level": self.log_level
                }
            }
        }

    def getlogger(self, type):
        return logging.getLogger(type)

    def request_logging(self, logger, request, req_json):
        try:
            log_info = {
                "method": request.method,
                "path": request.path,
                "ip": request.remote_addr,
                "headers": request.headers,
                "full_path": request.full_path,
                "mac:": req_json["mac"],
                "timestamp": req_json["timestamp"],
                "data_version": req_json["data_version"]
            }
            logger.info(log_info)
        except:
            logger.error("request error")


if __name__ == '__main__':
    log_handler = LogHandler(service='CardTesting')
    init_logger = log_handler.getlogger('INIT')
    init_logger.info("success log")
    
    log_handler.service = "set1"
    log_handler.log_level = "ERROR"
    log_handler.set_logging()
    init_logger = log_handler.getlogger('INIT')
    init_logger.error("success log")
    



import pandas as pd
import numpy as np
from sklearn.decomposition import PCA


def total_score_to_policy_score(total_score):
    return round(total_score * -2 + 1, 6)


def policy_score_to_total_score(x):
    return round(1 - (x - 1) / (-2), 2)


def pca_column_rank(pca_model, list_need_features):
    n_pcs = pca_model.components_.shape[0]
    most_important = [np.abs(pca_model.components_[i]).argmax()
                      for i in range(n_pcs)]

    initial_feature_names = list_need_features
    most_important_names = [
        initial_feature_names[most_important[i]] for i in range(n_pcs)]
    dic = {'PC{}'.format(i): most_important_names[i] for i in range(n_pcs)}
    df = pd.DataFrame(dic.items())
    df["ratio"] = pca_model.explained_variance_ratio_
    return df


def pca_to_ahp(pca_df):
    list1 = []
    record = None
    idex = 1
    for i in list(map(lambda x: str(x).split("0.")[1], list(pca_df.ratio.values))):
        for ii in i:
            if ii != "0":
                if record == ii or record is None:
                    list1.append(idex)
                else:
                    idex += 1
                    list1.append(idex)
                record = ii
                break
    dict_from_list = dict(zip(list(pca_df[1]), list1))
    return dict_from_list


if __name__ == '__main__':
    df = pd.read_csv("model_predict.csv")
    features = ['device_consistency_score', 'internet_info_score',
                'personal_device_score', 'device_connection_score', 'ip_change_score',
                'ip_connection_score', 'bio_behavior_score', 'robot_detection_score']
    x = df.loc[:, features].values
    model = PCA(n_components=8)
    model.fit(x)
    df1 = pca_column_rank(model, features)
    dict_from_list = pca_to_ahp(df1)
    print(dict_from_list)


import pandas as pd
pd.set_option('display.max_columns', None) 

type_data_analyse = {
    "path": "./data/",
    "df_analyze": "new_fraud_analyze.csv",
    "df_info": "new_tb_fraud_detect_info.csv",
    "threshold_police": 0.05
}


class Utils:
    def model_result_times(self, model_reason):
        normal = 0
        abnormal = 0
        for reason in model_reason:
            if "-normal" in reason:
                normal += 1
            else:
                abnormal += 1
        return normal, abnormal

    def model_typle(self, res):
        if res != "0":
            return 1
        else:
            return 0

    def count_rank_to_dict(dict_input_result, rank=4):
        top_keys = list(dict_input_result.keys())[0:rank]
        top_values = list(dict_input_result.values())[0:rank]
        
        dict_result = {}
        for k, v in zip(top_keys, top_values):
            dict_result[k] = v
        others_values = sum(list(dict_input_result.values())[rank:])
        dict_result["others"] = others_values
        return dict_result


class DataAnalyse(Utils):
    def __init__(self, type_data_analyse=type_data_analyse):
        self.type_data_analyse = type_data_analyse
        self.intersected_df = None
        self.intersected_df_normal = None
        self.df_all = None

    def read_fraud(self):
        df_ana = pd.read_csv(self.type_data_analyse['path'] + self.type_data_analyse['df_analyze'])
        df_info = pd.read_csv(self.type_data_analyse['path'] + self.type_data_analyse['df_info'])
        # print(df_ana.columns)

        abnormal_df = df_ana[df_ana["policy_score"] < self.type_data_analyse['threshold_police']]
        normal_df = df_ana[df_ana["policy_score"] >= self.type_data_analyse['threshold_police']]
        df_info = df_info.rename(columns={'pk_id': 'fd_info_id'})
        intersected_df = pd.merge(df_info, abnormal_df, on=['fd_info_id'], how='inner')
        intersected_df_normal = pd.merge(df_info, normal_df, on=['fd_info_id'], how='inner')
        self.intersected_df = intersected_df
        self.intersected_df_normal = intersected_df_normal

    def data_analyse(self):
        df_all = pd.DataFrame()
        for udid in self.intersected_df["udid"].value_counts().keys():
            df = self.intersected_df[self.intersected_df["udid"] == udid]

            fruit_dict = {
                "UDID": udid,
                "累積數量": len(df),
                "client_account": str(list(set(df.client_account))),
                "設備種類": str(list(set(df.hardware_device_type))),
                "IP": str(list(set(df.ip_request))),
                "是否為proxy": str(df.ip_is_proxy.value_counts().to_dict()),
                "是否為vpn": str(df.ip_is_vpn.value_counts().to_dict()),
                "是否為tor": str(df.ip_is_tor.value_counts().to_dict()),
                "瀏覽器種類": str(list(set(df.browser_name))),
                "異常模型數量": 0,
                "設備一致性正常次數": str(self.model_result_times(df.device_consistency_reason)[0]),
                "設備一致性異常次數": str(self.model_result_times(df.device_consistency_reason)[1]),
                "設備慣性正常次數": str(self.model_result_times(df.personal_device_reason)[0]),
                "設備慣性異常次數": str(self.model_result_times(df.personal_device_reason)[1]),
                "網路資訊正常次數": str(self.model_result_times(df.internet_info_reason)[0]),
                "網路資訊異常次數": str(self.model_result_times(df.internet_info_reason)[1]),
                "設備連線次數正常次數": str(self.model_result_times(df.device_connection_reason)[0]),
                "設備連線次數異常次數": str(self.model_result_times(df.device_connection_reason)[1]),
                "IP連線次數正常次數": str(self.model_result_times(df.ip_connection_reason)[0]),
                "IP連線次數異常次數": str(self.model_result_times(df.ip_connection_reason)[1]),
                "IP切換正常次數": str(self.model_result_times(df.ip_change_reason)[0]),
                "IP切換異常次數": str(self.model_result_times(df.ip_change_reason)[1]),
                "生物行為正常次數": str(self.model_result_times(df.bio_behavior_reason)[0]),
                "生物行為異常次數": str(self.model_result_times(df.bio_behavior_reason)[1]),
                "機器人特徵正常次數": str(self.model_result_times(df.robot_detection_reason)[0]),
                "機器人特徵異常次數": str(self.model_result_times(df.robot_detection_reason)[1]),
                "此UDID是否存在正常分佈中": str(udid in list(self.intersected_df_normal["udid"])),
            }

            dict_ab_model = self.model_typle(fruit_dict["設備一致性異常次數"]) + self.model_typle(fruit_dict["設備慣性異常次數"]) +\
                self.model_typle(fruit_dict["網路資訊異常次數"]) + self.model_typle(fruit_dict["設備連線次數異常次數"]) +\
                self.model_typle(fruit_dict["IP連線次數異常次數"]) + self.model_typle(fruit_dict["IP切換異常次數"]) +\
                self.model_typle(fruit_dict["生物行為異常次數"]) + \
                self.model_typle(fruit_dict["機器人特徵異常次數"])

            fruit_dict["異常模型數量"] = dict_ab_model

            df_all = df_all.append([fruit_dict])
        df_all.index = range(len(df_all))
        self.df_all = df_all


def main():
    da = DataAnalyse()
    da.read_fraud()
    da.data_analyse()
    da.df_all['UDID'] = da.df_all['UDID'].astype(str)
    with pd.ExcelWriter('Courses.xlsx') as writer:
        da.df_all.to_excel(writer, sheet_name='異常')
        da.intersected_df_normal.to_excel(writer, sheet_name='正常')


if __name__ == '__main__':
    main()
